﻿using System;
using System.ServiceModel;
using System.ServiceModel.Description;
using System.Timers;

namespace Comm.Wcf.Client
{
    /// <summary>
    ///   Communication class to service. This class contains all
    ///   functionalities to etablish a connection to the 
    ///   service (including watchdog functionalities).
    /// </summary>
    public class CServiceClient
    {
        #region Event Declarations

        /// <summary>
        ///   Service was connected
        /// </summary>
        public event EventHandler Connected;

        /// <summary>
        ///   Service was disconnected
        /// </summary>
        public event EventHandler Disconnected;

        /// <summary>
        ///   Event handler which will be thrown, when UpdateHelloWorld from server was called
        /// </summary>
        public event EventHandler OnUpdateHelloWorld;

        #endregion

        #region Fields

        private const int ConnectionRetryInterval = 5000;
        private const int ConnectionWatchdogInterval = 10000;
        private const int RoundTripInterval = 15000;
        private readonly Timer _connectionRetryTimer = new Timer();
        private readonly Timer _connectionWatchdogTimer = new Timer();
        private readonly Guid _id = Guid.NewGuid();
        private readonly Timer _roundTripTimer = new Timer();
        private readonly string _userName = "";
        private readonly string _userPassword = "";
        private bool _autoActivate;
        private InstanceContext _context;
        private CDataCallback _dataCallback;
        private DataClient _dataClient;
        private string _helloWorld = String.Empty;
        private bool _isConnected;
        private bool _roundTripConfirmed;

        #endregion

        #region Properties

        /// <summary>
        ///   List with history message. This list contains all history
        ///   messages.
        /// </summary>
        public String HelloWorld
        {
            get { return _helloWorld; }
        }

        /// <summary>
        ///   Information if connection to service is etablished
        /// </summary>
        public Boolean IsConnected
        {
            get { return _isConnected; }
        }

        #endregion

        #region Constructor/Destructors

        /// <summary>
        ///   CServiceClient constructor
        /// </summary>
        /// <param name="userName"> Windows User Name </param>
        /// <param name="userPassword"> Windows User Password </param>
        public CServiceClient(String userName, String userPassword)
        {
            _userName = userName;
            _userPassword = userPassword;

            _connectionRetryTimer.Enabled = false;
            _connectionRetryTimer.Elapsed += ConnectionRetryTimerElapsed;
            _connectionRetryTimer.Interval = ConnectionRetryInterval;
            _connectionWatchdogTimer.Enabled = false;
            _connectionWatchdogTimer.Elapsed += ConnectionWatchdogTimerElapsed;
            _connectionWatchdogTimer.Interval = ConnectionWatchdogInterval;
            _roundTripTimer.Enabled = false;
            _roundTripTimer.Elapsed += RoundTripTimerElapsed;
            _roundTripTimer.Interval = RoundTripInterval;
        }

        #endregion

        #region Methods

        /// <summary>
        ///   Open connection to service
        /// </summary>
        public void Open()
        {
            _autoActivate = true;
            ConnectToService();
        }

        /// <summary>
        ///   Request hello world content
        /// </summary>
        public void RequestHelloWorld()
        {
            if (_dataClient != null)
            {
                try
                {
                    _dataClient.RequestHelloWorld(_id);
                }
                catch (Exception)
                {
                    //Log message

                    DisconnectFromService(false);
                }
            }
        }

        private void ConnectToService()
        {
            try
            {
                //Create callback channel object
                _dataCallback = new CDataCallback();

                //Add event handlers for callback channel
                _dataCallback.OnUpdateHelloWorld += (DataCallback_OnUpdateHelloWorld);
                _dataCallback.OnAliveRoundTrip += DataCallbackOnAliveRoundTrip;

                //Create instance context for callback channel
                _context = new InstanceContext(_dataCallback);
                _context.Opening += DataCallbackOpening;
                _context.Opened += DataCallbackOpened;
                _context.Closing += DataCallbackClosing;

                //Create channel object
                _dataClient = new DataClient(_context);
                _dataClient.InnerChannel.Opening += DataClientOpening;
                _dataClient.InnerChannel.Opened += DataClientOpened;
                _dataClient.InnerChannel.Closing += DataClientClosing;

                ServiceEndpoint serviceEndpoint = _dataClient.Endpoint;
                EndpointAddress endpointAddress = serviceEndpoint.Address;
                Uri uri = endpointAddress.Uri;

                //Configure user credentials
                if (_userName != "")
                {
                    if (_dataClient.ClientCredentials != null)
                    {
                        _dataClient.ClientCredentials.Windows.ClientCredential.UserName = _userName;
                        _dataClient.ClientCredentials.Windows.ClientCredential.Password = _userPassword;
                    }
                }

                _connectionWatchdogTimer.Enabled = false;
                _roundTripTimer.Enabled = true;
                _connectionRetryTimer.Enabled = false;

                _isConnected = true;

                _dataClient.Subscribe(_id);

                _dataClient.RequestHelloWorld(_id);

                if (Connected != null)
                {
                    Connected(this, new EventArgs());
                }
            }
            catch (Exception)
            {
                //Log message

                DisconnectFromService(true);
            }
        }

        private void DisconnectFromService(bool forceDisconnect)
        {
            bool channelAlive = true;
            if (_isConnected || forceDisconnect)
            {
                _isConnected = false;
                try
                {
                    _dataClient.Unsubscribe(_id);
                }
                catch (Exception)
                {
                    channelAlive = false;
                }
                finally
                {
                    if (channelAlive)
                    {
                        //Log message
                    }
                    if (_dataCallback != null)
                    {
                        _dataCallback.OnAliveRoundTrip -= DataCallbackOnAliveRoundTrip;
                        _dataCallback.OnUpdateHelloWorld -= DataCallback_OnUpdateHelloWorld;

                        _dataCallback = null;
                    }

                    _dataClient = null;


                    _connectionWatchdogTimer.Enabled = false;
                    _roundTripTimer.Enabled = false;
                    _connectionRetryTimer.Enabled = _autoActivate;

                    if (Disconnected != null)
                    {
                        Disconnected(this, new EventArgs());
                    }
                }
            }
        }

        #endregion

        #region Event Handling

        private void ConnectionRetryTimerElapsed(object sender, ElapsedEventArgs e)
        {
            ConnectToService();
        }

        private void ConnectionWatchdogTimerElapsed(object sender, ElapsedEventArgs e)
        {
            _connectionWatchdogTimer.Enabled = false;

            if (_roundTripConfirmed == false)
            {
                DisconnectFromService(false);
            }
        }

        private void DataCallbackOnAliveRoundTrip(object sender, EventArgs e)
        {
            _roundTripConfirmed = true;
            _connectionWatchdogTimer.Enabled = false;
        }

        private void DataCallbackOpened(object sender, EventArgs e)
        {
            //Log message
        }

        private void DataCallbackOpening(object sender, EventArgs e)
        {
            //Log message
        }

        private void DataCallback_OnUpdateHelloWorld(object sender, EventArgs e)
        {
            _helloWorld = _dataCallback.HelloWorld;

            if (OnUpdateHelloWorld != null)
            {
                OnUpdateHelloWorld(this, new EventArgs());
            }
        }

        private void DataClientClosing(object sender, EventArgs e)
        {
            ((DataClient) sender).InnerChannel.Opening -= DataClientOpening;
            ((DataClient) sender).InnerChannel.Opened -= DataClientOpened;
            ((DataClient) sender).InnerChannel.Closing -= DataClientClosing;
        }

        private void DataClientOpened(object sender, EventArgs e)
        {
            //Log message
        }

        private void DataClientOpening(object sender, EventArgs e)
        {
            //Log message
        }

        private void DataCallbackClosing(object sender, EventArgs e)
        {
            ((InstanceContext) sender).Opening -= DataCallbackOpening;
            ((InstanceContext) sender).Opened -= DataCallbackOpened;
            ((InstanceContext) sender).Closing -= DataCallbackClosing;
        }

        private void RoundTripTimerElapsed(object sender, ElapsedEventArgs e)
        {
            _roundTripConfirmed = false;
            try
            {
                _dataClient.RequestAliveRoundTrip(_id);

                _connectionWatchdogTimer.Enabled = true;
            }
            catch (Exception)
            {
                DisconnectFromService(false);
            }
        }

        #endregion
    }
}