﻿using System;
using System.Configuration;
using System.Threading;

namespace Comm.Wcf.Client
{
    /// <summary>
    ///   Thread for WCF communication.
    /// </summary>
    public class CServiceThread
    {
        #region Event Declarations

        /// <summary>
        ///   Event handler when service connected
        /// </summary>
        public event EventHandler ServiceConnected;

        /// <summary>
        ///   Event handler when service disconnected
        /// </summary>
        public event EventHandler ServiceDisconnected;

        #endregion

        #region Fields

        private readonly CServiceClient _service;
        private readonly Thread _thread;

        private static CServiceThread _instance;

        private bool _notifyConnect;
        private bool _notifyDisconnect;
        private bool _serviceActive;

        #endregion

        #region Properties

        /// <summary>
        ///   WCF service
        /// </summary>
        public CServiceClient Service
        {
            get { return _service; }
        }

        /// <summary>
        ///   Activating WCF service
        /// </summary>
        public bool ServiceActive
        {
            set
            {
                if (_serviceActive != value)
                {
                    _serviceActive = value;
                    _notifyConnect = _serviceActive;
                    _notifyDisconnect = (_serviceActive == false);
                }
            }
            get { return _serviceActive; }
        }

        #endregion

        #region Constructor/Destructors

        /// <summary>
        ///   Constructor
        /// </summary>
        private CServiceThread()
        {
            _service = new CServiceClient(ConfigurationManager.AppSettings["UserName"],
                                          ConfigurationManager.AppSettings["UserPassword"]);

            _thread = new Thread(ThreadExecute) {IsBackground = true};
            _thread.Start();
        }

        /// <summary>
        ///   Destructor
        /// </summary>
        ~CServiceThread()
        {
            _thread.Abort();
            _thread.Join();
        }

        #endregion

        #region Methods

        /// <summary>
        ///   Provides the instance of the object. The object will be created, if the object does not exist (Singleton).
        /// </summary>
        /// <returns> Returns instance of the object. </returns>
        public static CServiceThread GetInstance()
        {
            //Create object, when CServiceThread object does not exist
            return _instance ?? (_instance = new CServiceThread());
            //Return instance of CServiceObject object
        }

        /// <summary>
        ///   Thread main execute method
        /// </summary>
        private void ThreadExecute()
        {
            const byte cSleepInterval = 50;

            _service.Connected += _service_Connected;
            _service.Disconnected += _service_Disconnected;

            while (true)
            {
                try
                {
                    if (_notifyConnect)
                    {
                        _service.Open();
                        _notifyConnect = false;
                    }
                    if (_notifyDisconnect)
                    {
                        _notifyDisconnect = false;
                    }

                    //Timeout Thread
                    Thread.Sleep(cSleepInterval);
                }
                catch (ThreadAbortException)
                {
                    break;
                }
            }
        }

        #endregion

        #region Event Handling

        private void _service_Connected(object sender, EventArgs e)
        {
            if (ServiceConnected != null)
            {
                ServiceConnected(this, new EventArgs());
            }
        }

        private void _service_Disconnected(object sender, EventArgs e)
        {
            if (ServiceDisconnected != null)
            {
                ServiceDisconnected(this, new EventArgs());
            }
        }

        #endregion
    }
}