﻿using System;
using System.Collections.Generic;
using System.Runtime.Serialization;
using System.ServiceModel;
using System.Threading;
using System.Timers;
using System.Windows.Forms;
using Comm.Wcf.Interfaces;
using Comm.Wcf.Utils;
using Timer = System.Timers.Timer;

namespace Comm.Wcf.Server
{
    /// <summary>
    ///   Service server class
    /// </summary>
    [DataContract]
    [ServiceBehavior(InstanceContextMode = InstanceContextMode.Single)]
    public class CServiceServer : IData
    {
        #region Event Declarations

        /// <summary>
        ///   Subscriber list has been changed
        /// </summary>
        public event EventHandler SubscriberListChanged;

        #endregion

        #region Fields

        private const int BoringWatchdogInterval = 30000;
        private readonly Timer _boringTimer = new Timer();
        private readonly Dictionary<Guid, CSubscriber> _subscribers = new Dictionary<Guid, CSubscriber>();

        #endregion

        #region Properties

        /// <summary>
        ///   Number of connected Clients
        /// </summary>
        public int ConnectedClients
        {
            get { return _subscribers.Count; }
        }

        #endregion

        #region Constructor/Destructors

        public CServiceServer()
        {
            _boringTimer.Elapsed += BoringTimerElapsed;
            _boringTimer.Interval = BoringWatchdogInterval;
            _boringTimer.Enabled = true;
        }

        #endregion

        #region Methods

        private void RemoveSubscriber(CSubscriber subscriber)
        {
            try
            {
                _subscribers.Remove(subscriber.Id);
                IDisposable disposable = subscriber;
                if (disposable != null)
                {
                    disposable.Dispose();
                }
            }
            catch (KeyNotFoundException)
            {
                //Log message
            }
            if (SubscriberListChanged != null)
            {
                SubscriberListChanged(this, new EventArgs());
            }
        }

        #endregion

        #region Event Handling

        private void BoringTimerElapsed(object sender, ElapsedEventArgs e)
        {
            try
            {
                //Request monitor
                Monitor.Enter(this);
                //BEGIN OF THREAD-SAFE-PART
                try
                {
                    foreach (KeyValuePair<Guid, CSubscriber> subscriber in _subscribers)
                    {
                        try
                        {
                            subscriber.Value.Callback.UpdateHelloWorld(String.Format("{0}, Time: {1}",
                                                                                     "Just to say it again: Hello world",
                                                                                     DateTime.Now.ToString()));
                        }
                        catch (Exception)
                        {
                            //Log message
                        }
                    }
                }
                catch (Exception)
                {
                    //Log message
                }
            }
            catch (Exception)
            {
                //Log message
            }
            finally
            {
                //END OF THREAD-SAFE-PART
                //Release monitor
                Monitor.Exit(this);
            }
        }

        private void SubscriberTimeout(object sender, EventArgs e)
        {
            ((CSubscriber) sender).Timeout -= SubscriberTimeout;
            Guid id = (sender as CSubscriber).Id;
            Unsubscribe(id);
        }

        #endregion

        #region IData Members

        /// <summary>
        ///   Subscribe client
        /// </summary>
        /// <param name="id"> Unique identification number </param>
        public void Subscribe(Guid id)
        {
            bool subscriberListChanged = false;

            try
            {
                //Request monitor
                Monitor.Enter(this);
                //BEGIN OF THREAD-SAFE-PART
                try
                {
                    CSubscriber subscriber = _subscribers[id];
                    subscriber.RefreshTimeout();

                    subscriber.Callback = OperationContext.Current.GetCallbackChannel<IDataCallback>();
                    Application.DoEvents();
                }
                catch (KeyNotFoundException)
                {
                    var subscriber = new CSubscriber();
                    subscriber.Timeout += SubscriberTimeout;
                    subscriber.Id = id;
                    try
                    {
                        _subscribers.Add(id, subscriber);
                        Application.DoEvents();
                        subscriber.Callback = OperationContext.Current.GetCallbackChannel<IDataCallback>();
                        Application.DoEvents();
                        subscriberListChanged = true;
                    }
                    catch (Exception)
                    {
                        RemoveSubscriber(subscriber);
                        subscriberListChanged = true;
                    }
                }
            }
            catch (Exception)
            {
                //Log message
            }
            finally
            {
                //END OF THREAD-SAFE-PART
                //Release monitor
                Monitor.Exit(this);
            }
            if (subscriberListChanged)
            {
                if (SubscriberListChanged != null)
                {
                    SubscriberListChanged(this, new EventArgs());
                }
            }
        }

        /// <summary>
        ///   Unsubscribe client
        /// </summary>
        /// <param name="id"> unique identification number </param>
        public void Unsubscribe(Guid id)
        {
            try
            {
                //Request monitor
                Monitor.Enter(this);
                //BEGIN OF THREAD-SAFE-PART
                try
                {
                    CSubscriber subscriber = _subscribers[id];
                    subscriber.Callback = null;
                    RemoveSubscriber(subscriber);
                    subscriber.Timeout -= SubscriberTimeout;
                    subscriber.Dispose();
                }
                catch (KeyNotFoundException)
                {
                    //Log message
                }
            }
            catch (Exception)
            {
                //Log message
            }
            finally
            {
                //END OF THREAD-SAFE-PART
                //Release monitor
                Monitor.Exit(this);
            }
        }

        void IData.RequestAliveRoundTrip(Guid id)
        {
            try
            {
                //Request monitor
                Monitor.Enter(this);
                //BEGIN OF THREAD-SAFE-PART
                try
                {
                    CSubscriber subscriber = _subscribers[id];
                    subscriber.RefreshTimeout();
                    try
                    {
                        subscriber.Callback.SetAliveRoundTrip();
                    }
                    catch (Exception)
                    {
                        RemoveSubscriber(subscriber);
                    }
                }
                catch (KeyNotFoundException)
                {
                    //Log message
                }
            }
            catch (Exception)
            {
                //Log message
            }
            finally
            {
                //END OF THREAD-SAFE-PART
                //Release monitor
                Monitor.Exit(this);
            }
        }


        void IData.RequestHelloWorld(Guid id)
        {
            try
            {
                //Request monitor
                Monitor.Enter(this);
                //BEGIN OF THREAD-SAFE-PART
                try
                {
                    CSubscriber subscriber = _subscribers[id];
                    subscriber.RefreshTimeout();
                    try
                    {
                        subscriber.Callback.UpdateHelloWorld(String.Format("{0}, Time: {1}", "Hello world",
                                                                           DateTime.Now.ToString()));
                    }
                    catch (Exception)
                    {
                        RemoveSubscriber(subscriber);
                    }
                }
                catch (KeyNotFoundException)
                {
                }
            }
            catch (Exception)
            {
                //Log message
            }
            finally
            {
                //END OF THREAD-SAFE-PART
                //Release monitor
                Monitor.Exit(this);
            }
        }

        #endregion
    }
}